import os
from pathlib import Path
from typing import List, Optional, Union

import click

from json_schema_for_humans.schema.schema_to_render import SchemaToRender


def _get_schema_paths(schema_file_or_dir: Union[str, Path]) -> List[Path]:
    schema_file_paths: List[Path] = []
    if isinstance(schema_file_or_dir, str):
        schema_file_or_dir = Path(schema_file_or_dir)

    if schema_file_or_dir.is_file() or (schema_file_or_dir.is_absolute() and not schema_file_or_dir.is_dir()):
        schema_file_paths.append(schema_file_or_dir)
    elif schema_file_or_dir.is_dir():
        for glob_pattern in ["**/*.json", "**/*.yaml", "**/*.yml"]:
            schema_file_paths += [schema_path for schema_path in schema_file_or_dir.glob(glob_pattern)]
    else:
        schema_file_paths += [schema_path for schema_path in Path.cwd().glob(str(schema_file_or_dir))]

    return [p.absolute() for p in schema_file_paths]


def get_schemas_to_render(
    schema_file_or_dir: Union[str, Path],
    result_file_or_dir: Optional[Union[str, Path]],
    result_extension: str,
) -> List[SchemaToRender]:
    """Build and return a list of SchemaToRender containing the list of input schema and output file that will be
    rendered in a run
    """
    to_render: List[SchemaToRender] = []

    schema_file_paths = _get_schema_paths(schema_file_or_dir)

    if not result_file_or_dir:
        return [SchemaToRender(schema_file_path, None, None) for schema_file_path in schema_file_paths]

    if isinstance(result_file_or_dir, str):
        result_file_or_dir = Path(result_file_or_dir)

    # Compute output_dir where additional files will be copied if necessary
    result_path_is_dir = Path(result_file_or_dir).is_dir()
    output_dir = result_file_or_dir if result_path_is_dir else result_file_or_dir.parent
    if not output_dir.exists():
        raise FileNotFoundError(f"{output_dir} not found")

    if len(schema_file_paths) > 1 and not result_path_is_dir:
        raise AssertionError("Several schemas to render, but output path is not a directory")

    for schema_file_path in schema_file_paths:
        if result_path_is_dir:
            this_schema_result_path = (
                result_file_or_dir / f"{os.path.splitext(schema_file_path.name)[0]}.{result_extension}"
            )
        else:
            this_schema_result_path = result_file_or_dir
        to_render.append(SchemaToRender(schema_file_path, this_schema_result_path, output_dir))
    return to_render


def get_result_output(
    output_path_or_file: Optional[Path], schema_files_or_dir: List[str], result_extension: str
) -> Path:
    if not output_path_or_file:
        if len(schema_files_or_dir) > 1 or any(Path(p).is_dir() for p in schema_files_or_dir):
            output_path_or_file = Path.cwd()
        else:
            output_path_or_file = Path(f"schema_doc.{result_extension}")

    if output_path_or_file.is_dir():
        if not output_path_or_file.exists():
            raise click.ClickException(f"Output path is a directory, but it does not exist: {output_path_or_file}")
    else:
        output_directory = output_path_or_file.parent
        if not output_directory.exists():
            raise click.ClickException(f"Output path file is in a directory that does not exist: {output_directory}")
    return output_path_or_file



from enum import Enum


class SchemaKeyword(Enum):
    REQUIRED = "required"
    TITLE = "title"
    CONTAINS = "contains"
    DESCRIPTION = "description"
    EXAMPLES = "examples"
    ITEMS = "items"
    PREFIX_ITEMS = "prefixItems"
    UNIQUE_ITEMS = "uniqueItems"
    ADDITIONAL_ITEMS = "additionalItems"
    MAX_ITEMS = "maxItems"
    MIN_ITEMS = "minItems"
    MAX_LENGTH = "maxLength"
    MIN_LENGTH = "minLength"
    MULTIPLE_OF = "multipleOf"
    MINIMUM = "minimum"
    EXCLUSIVE_MINIMUM = "exclusiveMinimum"
    MAXIMUM = "maximum"
    EXCLUSIVE_MAXIMUM = "exclusiveMaximum"
    PATTERN = "pattern"
    CONST = "const"
    ENUM = "enum"
    META_ENUM = "meta:enum"
    ELSE = "else"
    THEN = "then"
    IF = "if"
    NOT = "not"
    ONE_OF = "oneOf"
    ANY_OF = "anyOf"
    ALL_OF = "allOf"
    PROPERTIES = "properties"
    PATTERN_PROPERTIES = "patternProperties"
    ADDITIONAL_PROPERTIES = "additionalProperties"
    FORMAT = "format"
    TYPE = "type"



import json
import re
from datetime import datetime
from typing import Any, List, Optional, cast

import yaml
from jinja2 import Environment, pass_environment
from markdown2 import Markdown  # type: ignore
from markupsafe import Markup
from markupsafe import escape as markupsafe_escape
from pygments import highlight
from pygments.formatters.html import HtmlFormatter
from pygments.lexers.data import YamlLexer
from pygments.lexers.javascript import JavascriptLexer
from pytz import reference

from json_schema_for_humans import const
from json_schema_for_humans.generation_configuration import GenerationConfiguration
from json_schema_for_humans.schema.schema_node import SchemaNode
from json_schema_for_humans.templating_utils import schema_keyword_convert_to_str, schema_keyword_to_str

SHORT_DESCRIPTION_NUMBER_OF_LINES = 8
DEFAULT_PATTERN = r"(\[Default - `([^`]+)`\])"
DEPRECATED_MARKER = r"[Deprecated"


def is_combining(schema_node: SchemaNode) -> bool:
    """Test if a schema is one of the combining schema keyword"""
    return bool({"anyOf", "allOf", "oneOf", "not"}.intersection(schema_node.keywords.keys()))


def is_text_short(text: str) -> bool:
    """Check if a string is short so that we can decide whether to make the section containing it expandable or not.
    The heuristic is counting 1 for each line + 1 for each group of 80 characters a line has
    """
    return (
        sum((len(line) / const.LINE_WIDTH + 1) for line in str(text).splitlines()) < SHORT_DESCRIPTION_NUMBER_OF_LINES
    )


def is_deprecated(_schema: SchemaNode) -> bool:
    """Test. Check if a property is deprecated without looking in description"""
    return False


def is_deprecated_look_in_description(schema_node: SchemaNode) -> bool:
    """Test. Check if a property is deprecated looking in description"""
    if const.DESCRIPTION not in schema_node.keywords:
        return False

    return bool(DEPRECATED_MARKER in (schema_node.keywords[const.DESCRIPTION].literal_str or ""))


def get_required_properties(schema_node: SchemaNode) -> List[str]:
    required_properties_node: Optional[SchemaNode] = schema_node.keywords.get("required")
    required_properties: List[str] = []
    if required_properties_node:
        required_properties = [p.literal_str for p in required_properties_node.array_items if p.literal_str]

    return required_properties


def get_first_property(schema_node: SchemaNode) -> Any:
    """
    Filter. get first property of given schema no matter the property key
    Usage:
    md template does not recurse on schema to render the if portion
    instead it renders the if in the heading directly
    """
    properties = schema_node.properties
    if not properties:
        return None
    first_property_name = next(iter(properties))
    return properties[first_property_name]


def get_undocumented_required_properties(schema_node: SchemaNode) -> List[str]:
    """Get the name of the properties that are required but not documented with their own node"""
    return [prop for prop in get_required_properties(schema_node) if prop not in schema_node.properties.keys()]


def python_to_json(value: Any) -> Any:
    """Filter. Return the value as it needs to be displayed in JSON

    Used to display a string literals more explicitly for default and const values.
    """
    return json.dumps(value, indent=4, separators=(",", ": "), ensure_ascii=False)


@pass_environment
def get_description(env: Environment, schema_node: SchemaNode) -> str:
    """Filter. Get the description of a property or an empty string"""
    description = schema_node.description
    return get_description_literal(env, description)


@pass_environment
def get_description_literal(env: Environment, description: str) -> str:
    """Filter. Get the description of a property or an empty string"""

    config: GenerationConfiguration = cast(GenerationConfiguration, env.globals["jsfh_config"])
    if config.default_from_description:
        match = re.match(DEFAULT_PATTERN, description)
        if match:
            description = description[match.span(1)[1] :].lstrip()

    if description and config.description_is_markdown and not config.result_extension == "md":
        # Markdown templates are expected to already have Markdown descriptions
        md: Markdown = env.globals["jsfh_md"]
        description = Markup(md.convert(markupsafe_escape(description)))

    return description


def get_default(schema_node: SchemaNode) -> str:
    """Filter. Return the default value for a property"""
    return str(schema_node.default_value) if schema_node.default_value is not None else ""


def get_default_look_in_description(schema_node: SchemaNode) -> str:
    """Filter. Get the default value of a JSON Schema property. If not set, look for it in the description."""
    default_value = schema_node.default_value
    if default_value:
        return default_value

    description = schema_keyword_to_str(schema_node, const.DESCRIPTION)
    if not description:
        return ""

    match = re.match(DEFAULT_PATTERN, description)
    if not match:
        return ""

    return match.group(2)


def get_numeric_restrictions_text(schema_node: SchemaNode, before_value: str = "", after_value: str = "") -> str:
    """Filter. Get the text to display about restrictions on a numeric type(integer or number)"""
    multiple_of = schema_keyword_convert_to_str(schema_node, const.MULTIPLE_OF)
    maximum = schema_keyword_convert_to_str(schema_node, const.MAXIMUM)
    exclusive_maximum = schema_keyword_convert_to_str(schema_node, const.EXCLUSIVE_MAXIMUM)
    minimum = schema_keyword_convert_to_str(schema_node, const.MINIMUM)
    exclusive_minimum = schema_keyword_convert_to_str(schema_node, const.EXCLUSIVE_MINIMUM)

    # Fix minimum and exclusive_minimum both there
    if minimum is not None and exclusive_minimum is not None:
        if minimum <= exclusive_minimum:
            exclusive_minimum = None
        else:
            minimum = None

    minimum_fragment = ""
    if minimum is not None:
        minimum_fragment += f"greater or equal to {before_value}{minimum}{after_value}"
    if exclusive_minimum is not None:
        minimum_fragment += f"strictly greater than {before_value}{exclusive_minimum}{after_value}"

    # Fix maximum and exclusive_maximum both there
    if maximum is not None and exclusive_maximum is not None:
        if maximum > exclusive_maximum:
            exclusive_maximum = None
        else:
            maximum = None

    maximum_fragment = ""
    if maximum is not None:
        maximum_fragment += f"lesser or equal to {before_value}{maximum}{after_value}"
    if exclusive_maximum is not None:
        maximum_fragment += f"strictly lesser than {before_value}{exclusive_maximum}{after_value}"

    result = "Value must be "
    touched = False
    if minimum_fragment:
        touched = True
        result += minimum_fragment
    if maximum_fragment:
        if touched:
            result += " and "
        touched = True
        result += maximum_fragment
    if multiple_of:
        if touched:
            result += " and "
        result += f"a multiple of {before_value}{multiple_of}{after_value}"

    return result if touched else ""


def deprecated(config, schema: SchemaNode) -> bool:
    return is_deprecated_look_in_description(schema) if config.deprecated_from_description else is_deprecated(schema)


def first_line(example_text: str) -> str:
    """Filter. Retrieve first line of string + add ... at the end if text has multiple lines cut line at max_length"""
    lines = example_text.splitlines()
    result = lines[0]
    etc = len(lines) > 1
    return f"{result}{' ...' if etc else ''}"


def get_local_time() -> str:
    return datetime.now(tz=reference.LocalTimezone()).strftime("%Y-%m-%d at %H:%M:%S %z")


def highlight_json_example(example_text: str) -> str:
    """Filter. Return an highlighted version of the provided JSON text"""
    return highlight(example_text, JavascriptLexer(), HtmlFormatter())


def yaml_example(example_text: str) -> str:
    """Filter. Return a YAML version of the provided JSON text"""
    loaded_example = json.loads(example_text)
    if not isinstance(loaded_example, dict):
        # YAML dump does not like things that are not object
        return str(loaded_example)
    return yaml.dump(loaded_example, allow_unicode=True, sort_keys=False)


def highlight_yaml_example(example_text: str) -> str:
    """Filter. Return a highlighted YAML version of the provided JSON text"""
    return highlight(yaml_example(example_text), YamlLexer(), HtmlFormatter())



from dataclasses import dataclass
from io import FileIO
from pathlib import Path
from tempfile import _TemporaryFileWrapper
from typing import Any, Dict, Optional, TextIO, Union

from json_schema_for_humans.const import FileLikeType
from json_schema_for_humans.schema.intermediate_representation import build_intermediate_representation
from json_schema_for_humans.template_renderer import TemplateRenderer


class NoResultFileError(EnvironmentError):
    """Custom Exception for when there is no result file"""


@dataclass
class SchemaToRender:
    def __init__(
        self,
        schema_file: Union[Path, FileLikeType, _TemporaryFileWrapper],
        result_file: Optional[Union[Path, FileLikeType, _TemporaryFileWrapper]],
        output_dir: Optional[Path],
    ):
        self.schema_file = schema_file
        if isinstance(self.schema_file, Path):
            self.schema_file = self.schema_file.resolve().absolute()
        self.result_file = result_file
        if isinstance(self.result_file, Path):
            self.result_file = self.result_file.resolve().absolute()
        self.output_dir = output_dir
        if self.output_dir and isinstance(self.output_dir, Path):
            self.output_dir = self.output_dir.resolve().absolute()

    @property
    def should_write_to_disk(self) -> bool:
        return self.result_file is not None

    @property
    def schema_file_name(self) -> str:
        return self.schema_file.name

    @property
    def name_for_output(self) -> str:
        """Name of the result documentation used in log messages"""
        if self.result_file:
            return self.result_file.name
        return self.schema_file.name

    def render(self, template_renderer: TemplateRenderer, loaded_schemas: Optional[Dict[str, Any]] = None) -> str:
        """Render the schema. If a result file is defined, write to disk. Return the rendered documentation"""
        schema_file = (
            open(file=str(self.schema_file), encoding="utf-8")
            if isinstance(self.schema_file, Path)
            else self.schema_file
        )
        intermediate_schema = build_intermediate_representation(schema_file, template_renderer.config, loaded_schemas)
        rendered = template_renderer.render(intermediate_schema)

        if self.should_write_to_disk:
            self._write_to_disk(rendered)

        return rendered

    def _write_to_disk(self, rendered: str):
        if not self.result_file:
            raise NoResultFileError(
                "Attempt to write rendered schema documentation to disk, but no result file or path defined"
            )

        if isinstance(self.result_file, Path):
            with self.result_file.open("w", encoding="utf-8") as fp:
                fp.write(rendered)
        elif isinstance(self.result_file, FileIO):
            self.result_file.write(rendered.encode("utf-8"))
        else:
            # self.result_file is file-like object
            self.result_file.write(rendered)



import re
from typing import Optional

import htmlmin  # type: ignore  # No stubs available
import jinja2
import markdown2  # type: ignore  # No stubs available
from jinja2 import FileSystemLoader, Template
from jinja2.ext import loopcontrols

from json_schema_for_humans import jinja_filters, templating_utils
from json_schema_for_humans.generation_configuration import GenerationConfiguration
from json_schema_for_humans.md_template import MarkdownTemplate
from json_schema_for_humans.schema.schema_node import SchemaNode


def _minify(rendered: str, is_markdown: bool, is_html: bool) -> str:
    if is_markdown:
        return re.sub(r"\n\s*\n", "\n\n", rendered)
    if is_html:
        return htmlmin.minify(rendered)
    return rendered


class TemplateRenderer:
    def __init__(self, config: GenerationConfiguration, template: Optional[Template] = None):
        self.config = config
        self.template = template or self._get_jinja_template()

    def _get_jinja_template(self) -> Template:
        loader = FileSystemLoader(self.config.template_path.parent)
        env = jinja2.Environment(
            loader=loader,
            extensions=[loopcontrols],
            trim_blocks=(self.config.template_is_markdown),
            lstrip_blocks=(self.config.template_is_markdown),
        )
        env.globals["jsfh_config"] = self.config
        env.globals["jsfh_md"] = markdown2.Markdown(extras=self.config.markdown_options)
        if self.config.template_is_markdown:
            md_template = MarkdownTemplate(self.config)
            md_template.register_jinja(env)

        env.filters["python_to_json"] = jinja_filters.python_to_json
        env.filters["get_default"] = (
            jinja_filters.get_default_look_in_description
            if self.config.default_from_description
            else jinja_filters.get_default
        )
        env.filters["get_type_name"] = templating_utils.get_type_name
        env.filters["get_description"] = jinja_filters.get_description
        env.filters["get_description_literal"] = jinja_filters.get_description_literal
        env.filters["get_numeric_restrictions_text"] = jinja_filters.get_numeric_restrictions_text

        env.filters["get_required_properties"] = jinja_filters.get_required_properties
        env.filters["get_first_property"] = jinja_filters.get_first_property
        env.filters["get_undocumented_required_properties"] = jinja_filters.get_undocumented_required_properties
        env.filters["highlight_json_example"] = jinja_filters.highlight_json_example
        env.filters["highlight_yaml_example"] = jinja_filters.highlight_yaml_example
        env.filters["yaml_example"] = jinja_filters.yaml_example
        env.filters["first_line"] = jinja_filters.first_line

        env.tests["combining"] = jinja_filters.is_combining
        env.tests["description_short"] = jinja_filters.is_text_short
        env.tests["deprecated"] = lambda schema: jinja_filters.deprecated(self.config, schema)
        env.globals["examples_as_yaml"] = self.config.examples_as_yaml
        env.globals["get_local_time"] = jinja_filters.get_local_time

        with open(self.config.template_path, "r") as template_fp:
            template = env.from_string(template_fp.read())

        return template

    def render(self, intermediate_schema: SchemaNode) -> str:
        rendered = self.template.render(schema=intermediate_schema, config=self.config)

        if self.config.minify:
            rendered = _minify(rendered, self.config.template_is_markdown, self.config.template_is_html)

        return rendered
